﻿/*
 * Copyright 2020 faddenSoft
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
using System;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.Runtime.CompilerServices;

using PluginCommon;

namespace Abm {
    /// <summary>
    /// This visualizer is intended solely to display the circle data.
    /// </summary>
    public class VisAbmCircle : MarshalByRefObject, IPlugin, IPlugin_Visualizer {
        // IPlugin
        public string Identifier {
            get { return "ABM Circle Data Visualizer"; }
        }
        private IApplication mAppRef;
        private byte[] mFileData;

        // Visualization identifiers; DO NOT change or projects that use them will break.
        private const string VIS_ABM_CIRCLE = "abm-circle-data";

        private const string P_OFFSET = "offset";
        private const string P_COUNT = "count";
        private const string P_INCREMENT = "increment";

        // Visualization descriptors.
        private VisDescr[] mDescriptors = new VisDescr[] {
            new VisDescr(VIS_ABM_CIRCLE, "ABM Circle Data", VisDescr.VisType.Bitmap,
                new VisParamDescr[] {
                    new VisParamDescr("File offset (hex)",
                        P_OFFSET, typeof(int), 0, 0x00ffffff, VisParamDescr.SpecialMode.Offset, 0),
                    new VisParamDescr("Total pixels",
                        P_COUNT, typeof(int), 1, 256, 0, 80),
                    new VisParamDescr("Increment",
                        P_INCREMENT, typeof(int), 1, 256, 0, 8),
                })
        };


        // IPlugin
        public void Prepare(IApplication appRef, byte[] fileData, AddressTranslate addrTrans) {
            mAppRef = appRef;
            mFileData = fileData;
        }

        // IPlugin
        public void Unprepare() {
            mAppRef = null;
            mFileData = null;
        }

        // IPlugin_Visualizer
        public VisDescr[] GetVisGenDescrs() {
            if (mFileData == null) {
                return null;
            }
            return mDescriptors;
        }

        // IPlugin_Visualizer
        public IVisualization2d Generate2d(VisDescr descr,
                ReadOnlyDictionary<string, object> parms) {
            switch (descr.Ident) {
                case VIS_ABM_CIRCLE:
                    return GenerateBitmap(parms);
                default:
                    mAppRef.ReportError("Unknown ident " + descr.Ident);
                    return null;
            }
        }

        private IVisualization2d GenerateBitmap(ReadOnlyDictionary<string, object> parms) {
            int offset = Util.GetFromObjDict(parms, P_OFFSET, 0);
            int count = Util.GetFromObjDict(parms, P_COUNT, 0);
            int incr = Util.GetFromObjDict(parms, P_INCREMENT, 0);

            if (offset < 0 || offset >= mFileData.Length ||
                    count < 1 || count > 256 || incr < 1 || incr > 256) {
                mAppRef.ReportError("Invalid parameter");
                return null;
            }

            if (incr > count) {
                // Could throw an error.  Easier just to fix it.
                incr = count;
            }

            // The table is 512 bytes, divided into two 256-byte sections.  At the very least
            // we need 256+count bytes.
            if (offset + 512 > mFileData.Length) {
                mAppRef.ReportError("Data extends off end of file");
                return null;
            }

            // Find max offset (circle radius).
            int maxX = -1;
            int maxY = -1;
            for (int i = 0; i < count; i++) {
                byte xoff = mFileData[offset + i];
                if (maxX < xoff) {
                    maxX = xoff;
                }
                byte yoff = mFileData[offset + 256 + i];
                if (maxY < yoff) {
                    maxY = yoff;
                }
            }

            // Define a bitmap big enough to cover the circle.  Each horizontal item
            // represents two pixels.
            int width = (VisBitmap8.CharWidth + 1) * (maxX * 2 * 2 + 2);
            mAppRef.DebugLog("max is " + maxX + ", width=" + width);
            int height = (VisBitmap8.CharHeight + 1) * (maxY * 2 + 1);
            int centerX = maxX * 2;
            int centerY = maxY;

            VisBitmap8 vb = new VisBitmap8(width, height);
            SetPalette(vb);

            int colorIndex = FIRST_COLOR;
            int colorCtr = incr;
            for (int i = 0; i < count; i++) {
                int xoff = mFileData[offset + i];
                int yoff = mFileData[offset + 256 + i];

                DrawValue(vb, i, centerX + xoff * 2, centerY + yoff,
                    COLOR_BLACK, colorIndex + NUM_COLORS);
                DrawValue(vb, i, centerX - xoff * 2, centerY + yoff,
                    COLOR_BLACK, colorIndex + NUM_COLORS);
                DrawValue(vb, i, centerX + xoff * 2, centerY - yoff,
                    COLOR_BLACK, colorIndex + NUM_COLORS);
                DrawValue(vb, i, centerX - xoff * 2, centerY - yoff,
                    COLOR_BLACK, colorIndex);

                if (--colorCtr == 0) {
                    colorCtr = incr;
                    colorIndex++;
                    if (colorIndex == FIRST_COLOR + NUM_COLORS) {
                        colorIndex = FIRST_COLOR;
                    }
                }
            }

            return vb;
        }

        private static readonly char[] HEX_CHAR = {
            '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
        };
        private static void DrawValue(VisBitmap8 vb, int val, int cellX, int cellY,
                    int fgColor, int bgColor) {
            // Fill in left edge with background color.
            VisBitmap8.DrawChar(vb, ' ', cellX * (VisBitmap8.CharWidth + 1),
                cellY * (VisBitmap8.CharHeight + 1), (byte)bgColor, (byte)bgColor);

            char ch0 = HEX_CHAR[(val >> 4) & 0x0f];
            char ch1 = HEX_CHAR[val & 0x0f];
            VisBitmap8.DrawChar(vb, ch0, cellX * (VisBitmap8.CharWidth + 1) + 1,
                cellY * (VisBitmap8.CharHeight + 1), (byte)fgColor, (byte)bgColor);
            VisBitmap8.DrawChar(vb, ch1, (cellX + 1) * (VisBitmap8.CharWidth + 1),
                cellY * (VisBitmap8.CharHeight + 1), (byte)fgColor, (byte)bgColor);
        }

        const int FIRST_COLOR = 3;
        const int NUM_COLORS = 6;
        const int COLOR_BLACK = 1;
        const int COLOR_WHITE = 2;
        private void SetPalette(VisBitmap8 vb) {
            vb.AddColor(0xff, 0x40, 0x40, 0x40);    // 0=dark grey (bitmap background)
            vb.AddColor(0xff, 0x00, 0x00, 0x00);    // 1=black
            vb.AddColor(0xff, 0xff, 0xff, 0xff);    // 2=white

            int orangeRed = Util.MakeARGB(0xff, 0xdc, 0x14, 0x3c);
            int lawnGreen = Util.MakeARGB(0xff, 0x7c, 0xcf, 0x00);
            int deepSkyBlue = Util.MakeARGB(0xff, 0x00, 0xbf, 0xff);
            int tomato = Util.MakeARGB(0xff, 0xff, 0x63, 0x47);
            int springGreen = Util.MakeARGB(0xff, 0x00, 0xff, 0x7f);
            int mediumTurquoise = Util.MakeARGB(0xff, 0x48, 0xd1, 0xcc);

            // six bright colors
            vb.AddColor(orangeRed);
            vb.AddColor(lawnGreen);
            vb.AddColor(deepSkyBlue);
            vb.AddColor(tomato);
            vb.AddColor(springGreen);
            vb.AddColor(mediumTurquoise);

            // darker versions of the bright colors
            const float darkFactor = 0.6f;
            vb.AddColor(MultColor(orangeRed, darkFactor));
            vb.AddColor(MultColor(lawnGreen, darkFactor));
            vb.AddColor(MultColor(deepSkyBlue, darkFactor));
            vb.AddColor(MultColor(tomato, darkFactor));
            vb.AddColor(MultColor(springGreen, darkFactor));
            vb.AddColor(MultColor(mediumTurquoise, darkFactor));
        }

        private int MultColor(int color, float mult) {
            int newA = (color >> 24) & 0xff;
            int newR = (int)(((color >> 16) & 0xff) * mult);
            int newG = (int)(((color >> 8) & 0xff) * mult);
            int newB = (int)((color & 0xff) * mult);
            int res = (newA << 24) | (newR << 16) | (newG << 8) | newB;
            //mAppRef.DebugLog("Converted " + color.ToString("x8") + " to " + res.ToString("x8"));
            return res;
        }
    }
}
